# -*- coding: utf-8 -*-
#
#  bln.py - a easyballoon compatible Saori module for ninix
#  Copyright (C) 2002-2010 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

# TODO:
# - font.face

import os
import random
import math
import time
import sys

if 'DISPLAY' in os.environ:
    if 'gtk' not in sys.modules:
        try:
            import pygtk
            pygtk.require('2.0')
        except ImportError:
            pass
    import gtk
    import gobject
    import pango
    import ninix.pix
else:
    gtk = None

import ninix.script
from ninix.dll import SAORI


class Saori(SAORI):

    def __init__(self):
        SAORI.__init__(self)
        self.blns = {}
        self.__sakura = None

    def need_ghost_backdoor(self, sakura):
        self.__sakura = sakura

    def check_import(self):
        return 1 if self.__sakura and gtk else 0

    def setup(self):
        self.blns = self.read_bln_txt(self.dir)
        return 1 if self.blns else 0

    def read_bln_txt(self, dir):
        blns = {}
        try:
            f = open(os.path.join(dir, 'bln.txt'), 'r')
            data = {}
            name = ''
            for line in f:
                line = line.strip()
                if not line:
                    continue
                if line.startswith('//'):
                    continue
                if '[' in line:
                    if name:
                        blns[name] = (data, {})
                    data = {}
                    start = line.find('[')
                    end = line.find(']')
                    if end < 0:
                        end = len(line)
                    name = line[start + 1:end]
                else:
                    if ',' in line:
                        key, value = [x.strip() for x in line.split(',', 1)]
                        data[key] = value
            if name:
                blns[name] = (data, {})
            return blns
        except:
            return None

    def finalize(self):
        for name in self.blns:
            data, bln = self.blns[name]
            for bln_id in bln.keys():
                if bln[bln_id]:
                    bln[bln_id].destroy()
                    del bln[bln_id]
        self.blns = {}
        return 1

    def execute(self, argument):
        if not argument:
            return self.RESPONSE[400]
        name = argument[0]
        text = argument[1] if len(argument) >= 2 else ''
        if len(argument) >= 3:
            try:
                offset_x = int(argument[2])
            except:
                offset_x = 0
        else:
            offset_x = 0
        if len(argument) >= 4:
            try:
                offset_y = int(argument[3])
            except:
                offset_y = 0
        else:
            offset_y = 0
        bln_id = argument[4] if len(argument) >= 5 else ''
        if len(argument) >= 6 and argument[5] in ['1', '2']:
            update = int(argument[5])
        else:
            update = 0
        if name in self.blns:
            data, bln = self.blns[name]
            if bln_id in bln and update == 0:
                bln[bln_id].destroy()
                del bln[bln_id]
            if text:
                if update == 0 or bln_id not in bln:
                    bln[bln_id] = Balloon(self.__sakura, self.dir,
                                          data, text,
                                          offset_x, offset_y, name, bln_id)
                else:
                    bln[bln_id].update_script(text, update)
            self.blns[name] = (data, bln)
        elif name == 'clear':
            for name in self.blns:
                data, bln = self.blns[name]
                if bln_id in bln and \
                   bln[bln_id].get_state() != 'orusuban':
                    bln[bln_id].destroy()
                    del bln[bln_id]
                    self.blns[name] = (data, {})
        return None

class Balloon:

    def __init__(self, sakura, dir, data,
                 text, offset_x, offset_y, name, bln_id):
        self.dir = dir
        self.__sakura = sakura
        self.name = name
        self.id = bln_id
        self.data = data # XXX
        self.window = self.create_window()
        self.position = data.get('position', 'sakura')
        left, top, scrn_w, scrn_h = ninix.pix.get_workarea()
        # -1: left, 1: right
        if self.position == 'sakura':
            s0_x, s0_y, s0_w, s0_h = self.get_sakura_status('SurfaceSakura')
            self.direction = -1 if s0_x + s0_w / 2 > left + scrn_w / 2 else 1
        elif self.position == 'kero':
            s1_x, s1_y, s1_w, s1_h = self.get_sakura_status('SurfaceKero')
            self.direction = -1 if s1_x + s1_w / 2 > left + scrn_w / 2 else 1
        else:
            self.direction = 1 # XXX
        if self.position in ['sakura', 'kero']:
            if self.direction == -1:
                skin = data.get('skin.left', data.get('skin'))
            else:
                skin = data.get('skin.right', data.get('skin'))
        else:
            skin = data.get('skin')
        if skin is None:
            self.destroy()
            return
        path = os.path.join(self.dir, skin.replace('\\', '/'))
        try:
            balloon_pixbuf = ninix.pix.create_pixbuf_from_file(path)
        except:
            self.destroy()
            return
        self.scale = self.get_sakura_status('SurfaceScale') ## FIXME
        w = balloon_pixbuf.get_width()
        h = balloon_pixbuf.get_height()
        w = max(8, int(w * self.scale / 100))
        h = max(8, int(h * self.scale / 100))
        balloon_pixbuf = balloon_pixbuf.scale_simple(
            w, h, gtk.gdk.INTERP_BILINEAR)
        balloon_pixmap, mask_pixmap = balloon_pixbuf.render_pixmap_and_mask(1)
        self.x, self.y = self.get_coordinate(w, h)
        if 'offset.x' in data:
            self.x += self.direction * int(int(data['offset.x']) * self.scale / 100)
        if 'offset.y' in data:
            self.y += int(int(data['offset.y']) * self.scale / 100)
        if 'offset.random' in data:
            self.x += int(int(data['offset.random']) * random.randrange(-1, 2) * self.scale / 100)
            self.y += int(int(data['offset.random']) * random.randrange(-1, 2) * self.scale / 100)
        self.x += self.direction * int(offset_x * self.scale / 100)
        self.y += int(offset_y * self.scale / 100)
        self.window.move(self.x, self.y)
        gtk.gdk.flush() # XXX
        self.left = int(int(data.get('disparea.left', 0)) * self.scale / 100)
        self.right = int(int(data.get('disparea.right', w)) * self.scale / 100)
        self.top = int(int(data.get('disparea.top', 0)) * self.scale / 100)
        self.bottom = int(int(data.get('disparea.bottom', h)) * self.scale / 100)
        self.window.shape_combine_mask(mask_pixmap, 0, 0)
        self.script = None
        self.darea = gtk.DrawingArea()
        self.darea.set_events(gtk.gdk.EXPOSURE_MASK)
        self.darea.connect('expose_event', self.redraw)
        self.darea.set_size_request(w, h)
        self.darea.show()
        self.window.add(self.darea)
        self.darea.realize()
        self.surface_gc = self.darea.window.new_gc()
        self.darea.window.set_back_pixmap(balloon_pixmap, False)
        self.layout = None
        if text != 'noscript' and \
           (self.right - self.left) and (self.bottom - self.top):
            self.script = text
            if 'font.color' in data:
                fontcolor_r = int(data['font.color'][:2], 16)
                fontcolor_g = int(data['font.color'][2:4], 16)
                fontcolor_b = int(data['font.color'][4:6], 16)
                cmap = self.darea.get_colormap()
                self.surface_gc.foreground = cmap.alloc_color(
                    '#%02x%02x%02x' % (fontcolor_r, fontcolor_g, fontcolor_b))
            default_font_size = 12 # for Windows environment
            font_size = int(data.get('font.size', default_font_size))
            font_size = font_size * 3 / 4 # convert from Windows to GTK+
            font_size = font_size * self.scale / 100 * pango.SCALE
            self.layout = pango.Layout(self.darea.get_pango_context())
            self.font_desc = pango.FontDescription()
            self.font_desc.set_family('Sans') # FIXME
            self.font_desc.set_size(font_size)
            if data.get('font.bold') == 'on':
                self.font_desc.set_weight(pango.WEIGHT_BOLD)
            self.layout.set_font_description(self.font_desc)
            self.layout.set_wrap(pango.WRAP_CHAR)
            self.layout.set_width((self.right - self.left) * 1024)
        self.slide_vx = int(int(data.get('slide.vx', 0)) * self.scale / 100)
        self.slide_vy = int(int(data.get('slide.vy', 0)) * self.scale / 100)
        self.slide_autostop = int(data.get('slide.autostop', 0))
        if data.get('action.method') in ['sinwave', 'vibrate']:
            action = data['action.method']
            ref0 = int(data.get('action.reference0', 0))
            ref1 = int(data.get('action.reference1', 0))
            ref2 = int(data.get('action.reference2', 0))
            ref3 = int(data.get('action.reference3', 0))
            if ref2:
                self.action = {'method': action,
                               'ref0': ref0,
                               'ref1': ref1,
                               'ref2': ref2,
                               'ref3': ref3}
            else:
                self.action = None
        else:
            self.action = None
        self.move_notify_time = None
        self.life_time = None
        self.state = ''
        life = data.get('life', 'auto')
        if life == 'auto':
            self.life_time = 16000
        elif life in ['infinitely', '0']:
            pass
        elif life == 'orusuban':
            self.state = 'orusuban'
        else:
            try:
                self.life_time = int(life)
            except:
                pass
        self.start_time = time.time()
        self.startdelay = int(data.get('startdelay', 0))
        self.nooverlap = int('nooverlap' in data)
        self.talking = self.get_sakura_is_talking()
        self.move_notify_time = time.time()
        self.timeout_id = None
        self.visible = 0
        self.x_root = None
        self.y_root = None
        self.action_x = 0
        self.action_y = 0
        self.vx = 0
        self.vy = 0
        self.text_pos_x = self.left
        self.text_pos_y = self.top
        self.processed_script = None
        self.processed_text = ''
        self.text = ''
        self.script_wait = None
        self.quick_session = 0
        self.script_parser = ninix.script.Parser(error='loose')
        try:
            self.processed_script = self.script_parser.parse(self.script)
        except ninix.script.ParserError, e:
            self.processed_script = None
            print '-' * 50
            print e
            print self.script                                    
        self.timeout_id = gobject.timeout_add(10, self.do_idle_tasks)

    def clickerase(self):
        return self.data.get('clickerase', 'on') == 'on'

    def dragmove_horizontal(self):
        return self.data.get('dragmove.horizontal') == 'on'

    def dragmove_vertical(self):
        return self.data.get('dragmove.vertical') == 'on'

    def create_window(self):
        window = gtk.Window(type=gtk.WINDOW_POPUP)
        window.connect('delete_event', self.delete)
        window.connect('button_press_event', self.button_press)
        window.connect('button_release_event', self.button_release)
        window.connect('motion_notify_event', self.motion_notify)
        window.connect('leave_notify_event', self.leave_notify)
        window.set_events(gtk.gdk.BUTTON_PRESS_MASK|
                          gtk.gdk.BUTTON_RELEASE_MASK|
                          gtk.gdk.POINTER_MOTION_MASK|
                          gtk.gdk.LEAVE_NOTIFY_MASK)
        return window

    def get_coordinate(self, w, h):
        left, top, scrn_w, scrn_h = ninix.pix.get_workarea()
        s0_x, s0_y, s0_w, s0_h = self.get_sakura_status('SurfaceSakura')
        s1_x, s1_y, s1_w, s1_h = self.get_sakura_status('SurfaceKero')
        b0_x, b0_y, b0_w, b0_h = self.get_sakura_status('BalloonSakura')
        b1_x, b1_y, b1_w, b1_h = self.get_sakura_status('BalloonKero')
        x = left
        y = top
        if self.position == 'lefttop':
            pass
        elif self.position == 'leftbottom':
            y = top + scrn_h - h
        elif self.position == 'righttop':
            x = left + scrn_w - w
        elif self.position == 'rightbottom':
            x = left + scrn_w - w
            y = top + scrn_h - h
        elif self.position == 'center':
            x = left + (scrn_w - w) / 2
            y = top + (scrn_h - h) / 2
        elif self.position == 'leftcenter':
            y = top + (scrn_h - h) / 2
        elif self.position == 'rightcenter':
            x = left + scrn_w - w
            y = top + (scrn_h - h) / 2
        elif self.position == 'centertop':
            x = left + (scrn_w - w) / 2
        elif self.position == 'centerbottom':
            x = left + (scrn_w - w) / 2
            y = top + scrn_h - h
        elif self.position == 'sakura':
            if self.direction: # right
                x = s0_x + s0_w
            else:
                x = s0_x - w
            y = s0_y
        elif self.position == 'kero':
            if self.direction: # right
                x = s1_x + s1_w
            else:
                x = s1_x - w
            y = s1_y
        elif self.position == 'sakurab':
            x = b0_x
            y = b0_y
        elif self.position == 'kerob':
            x = b1_x
            y = b1_y
        return x, y

    def update_script(self, text, mode):
        if not text:
            return
        if mode == 2 and self.script is not None:
            self.script = ''.join((self.script, text))
        else:
            self.script = text
        self.processed_script = None
        self.processed_text = ''
        self.text = ''
        self.script_wait = None
        self.quick_session = 0
        try:
            self.processed_script = self.script_parser.parse(self.script)
        except ninix.script.ParserError, e:
            self.processed_script = None
            print '-' * 50
            print e
            print self.script                                    

    def get_sakura_is_talking(self):
        talking = 0
        try:
            talking = int(self.__sakura.is_talking())
        except:
            pass
        return talking

    def get_sakura_status(self, key):
        if key == 'SurfaceScale':
            result = self.__sakura.get_surface_scale()
        elif key == 'SurfaceSakura_Shown':
            result = int(self.__sakura.surface_is_shown(0))
        elif key == 'SurfaceSakura':
            try:
                s0_x, s0_y = self.__sakura.get_surface_position(0)
                s0_w, s0_h = self.__sakura.get_surface_size(0)
            except:
                s0_x, s0_y = 0, 0
                s0_w, s0_h = 0, 0
            result = s0_x, s0_y, s0_w, s0_h
        elif key == 'SurfaceKero_Shown':
            result = int(self.__sakura.surface_is_shown(1))
        elif key == 'SurfaceKero':
            try:
                s1_x, s1_y = self.__sakura.get_surface_position(1)
                s1_w, s1_h = self.__sakura.get_surface_size(1)
            except:
                s1_x, s1_y = 0, 0
                s1_w, s1_h = 0, 0
            result = s1_x, s1_y, s1_w, s1_h
        elif key == 'BalloonSakura_Shown':
            result = int(self.__sakura.balloon_is_shown(0))
        elif key == 'BalloonSakura':
            try:
                b0_x, b0_y = self.__sakura.get_balloon_position(0)
                b0_w, b0_h = self.__sakura.get_balloon_size(0)
            except:
                b0_x, b0_y = 0, 0
                b0_w, b0_h = 0, 0
            result = b0_x, b0_y, b0_w, b0_h
        elif key == 'BalloonKero_Shown':
            result = int(self.__sakura.balloon_is_shown(1))
        elif key == 'BalloonKero':
            try:
                b1_x, b1_y = self.__sakura.get_balloon_position(1)
                b1_w, b1_h = self.__sakura.get_balloon_size(1)
            except:
                b1_x, b1_y = 0, 0
                b1_w, b1_h = 0, 0
            result = b1_x, b1_y, b1_w, b1_h
        else:
            result = None
        return result

    def do_idle_tasks(self):
        if not self.window:
            return None
        s0_shown = self.get_sakura_status('SurfaceSakura_Shown')
        s1_shown = self.get_sakura_status('SurfaceKero_Shown')
        b0_shown = self.get_sakura_status('BalloonSakura_Shown')
        b1_shown = self.get_sakura_status('BalloonKero_Shown')
        sakura_talking = self.get_sakura_is_talking()
        if self.state == 'orusuban':
            if self.visible:
                if s0_shown or s1_shown:
                    self.destroy()
                    return None
            else:
                if not s0_shown and not s1_shown:
                    self.start_time = time.time()
                    self.visible = 1
                    self.window.show()
                    self.life_time = 300000
        else:
            if self.visible:
                if (self.position == 'sakura' and not s0_shown) or \
                        (self.position == 'kero' and not s1_shown) or \
                        (self.position == 'sakurab' and not b0_shown) or \
                        (self.position == 'kerob' and not b1_shown) or \
                        (self.nooverlap and not self.talking and sakura_talking):
                    self.destroy()
                    return None
            else:
                if time.time() - self.start_time >= self.startdelay * 0.001:
                    self.start_time = time.time()
                    self.visible = 1
                    self.window.show()
        if self.visible:
            if self.life_time:
                if time.time() - self.start_time >= self.life_time * 0.001 and \
                   not (self.processed_script or self.processed_text):
                    self.destroy()
                    return None
            if self.action:
                if  self.action['method'] == 'sinwave':
                    offset = self.action['ref1'] \
                             * math.sin(2.0 * math.pi
                                        * float(int((time.time() - \
                                                     self.start_time) * 1000) \
                                                % self.action['ref2'])
                                        / self.action['ref2'])
                    if self.action['ref0']:
                        self.action_y = int(offset * self.scale / 100)
                    else:
                        self.action_x = int(offset * self.scale / 100)
                elif self.action['method'] == 'vibrate':
                    offset = (int((time.time() - self.start_time) * 1000) / \
                              self.action['ref2']) % 2
                    self.action_x = int(offset * self.action['ref0'] * self.scale / 100)
                    self.action_y = int(offset * self.action['ref1'] * self.scale / 100)
            if (self.slide_vx != 0 or self.slide_vy != 0) and \
                   self.slide_autostop > 0 and \
                   self.slide_autostop * 0.001 + 0.05 <= time.time() - self.start_time:
                self.vx = self.direction * int((self.slide_autostop / 50.0 + 1) * self.slide_vx * self.scale / 100)
                self.slide_vx = 0
                self.vy = int((self.slide_autostop / 50.0 + 1) * self.slide_vy * self.scale / 100)
                self.slide_vy = 0
            if self.slide_vx != 0:
                self.vx = self.direction * int(((time.time() - self.start_time) * self.slide_vx * self.scale / 100) / 50 * 1000.)
            if self.slide_vy != 0:
                self.vy = int(((time.time() - self.start_time) * self.slide_vy * self.scale / 100) / 50 * 1000.)
            self.window.move(int(self.x + self.action_x + self.vx),
                             int(self.y + self.action_y + self.vy))
            if self.processed_script or self.processed_text:
                self.interpret_script()
        self.talking = 0 if self.talking and not sakura_talking else sakura_talking
        return True

    def redraw(self, widget, event):
        x, y, w, h = event.area
        self.darea.window.clear()
        if self.layout:
            self.darea.window.draw_layout(self.surface_gc,
                                          self.left, self.top, self.layout)

    def get_state(self):
        return self.state

    def interpret_script(self):
        if self.script_wait is not None:
            if time.time() < self.script_wait:
                return
            self.script_wait = None
        if self.processed_text:
            if self.quick_session or self.state == 'orusuban':
                self.text = ''.join((self.text, self.processed_text))
                self.draw_text(self.text)
                self.processed_text = ''
            else:
                self.text = ''.join((self.text, self.processed_text[0]))
                self.draw_text(self.text)
                self.processed_text = self.processed_text[1:]
                self.script_wait = time.time() + 0.014
            return
        node = self.processed_script.pop(0)
        if node[0] == ninix.script.SCRIPT_TAG:
            name, args = node[1], node[2:]
            if name == r'\n':
                self.text = ''.join((self.text, '\n'))
                self.draw_text(self.text)
            elif name == r'\w':
                if args:
                    try:
                        amount = int(args[0]) * 0.05 - 0.01
                    except ValueError:
                        amount = 0
                else:
                    amount = 1 * 0.05 - 0.01
                if amount > 0:
                    self.script_wait = time.time() + amount
            elif name == r'\b':
                if args:
                    try:
                        amount = int(args[0])
                    except ValueError:
                        amount = 0
                else:
                    amount = 1
                if amount > 0:
                    self.text = self.text[:-amount]
            elif name == r'\c':
                self.text = ''
            elif name == r'\_q':
                self.quick_session = not self.quick_session
            elif name == r'\l':
                self.life_time = None
                self.update_script('', 2)
        elif node[0] == ninix.script.SCRIPT_TEXT:
            text = ''
            for chunk in node[1]:
                text = ''.join((text, chunk[1]))
            self.processed_text = text

    def draw_text(self, text):
        self.layout.set_text(text)
        self.darea.queue_draw()

    def button_press(self, window, event):
        self.x_root = event.x_root
        self.y_root = event.y_root
        if event.type == gtk.gdk._2BUTTON_PRESS:
            x = int(event.x * 100 / self.scale)
            y = int(event.y * 100 / self.scale)
            self.__sakura.notify_event('OnEBMouseDoubleClick', self.name,
                                       x, y, self.id)
        return True

    def button_release(self, window, event):
        self.x_root = None
        self.y_root = None
        x = int(event.x * 100 / self.scale)
        y = int(event.y * 100 / self.scale)
        if event.type == gtk.gdk.BUTTON_RELEASE:
            if event.button == 1:
                self.__sakura.notify_event('OnEBMouseClick', self.name,
                                           x, y, self.id, 0)
            elif event.button == 3:
                self.__sakura.notify_event('OnEBMouseClick', self.name,
                                           x, y, self.id, 1)
        if self.clickerase():
            self.destroy()
        return True

    def motion_notify(self, window, event):
        if self.x_root is not None and \
           self.y_root is not None:
            x_delta = int(event.x_root - self.x_root)
            y_delta = int(event.y_root - self.y_root)
            if event.state & gtk.gdk.BUTTON1_MASK:
                if self.dragmove_horizontal():
                    self.x += x_delta
                if self.dragmove_vertical():
                    self.y += y_delta
                self.window.move(int(self.x + self.action_x + self.vx),
                                 int(self.y + self.action_y + self.vy))
                self.x_root = event.x_root
                self.y_root = event.y_root
        if self.move_notify_time is None or \
           time.time() - self.move_notify_time > 500 * 0.001:
            x = int(event.x * 100 / self.scale)
            y = int(event.y * 100 / self.scale)
            self.__sakura.notify_event('OnEBMouseMove', self.name,
                                       x, y, self.id)
            self.move_notify_time = time.time()
        return True

    def leave_notify(self, window, event):
        self.move_notify_time = None

    def delete(self, window, event):
        return True

    def destroy(self):
        self.visible = 0
        if self.window:
            self.window.destroy()
            self.window = None
        if self.timeout_id:
            gobject.source_remove(self.timeout_id)
            self.timeout_id = None
